//! Differnt types of Query functions (part of the Query builder API)
//!
//! Most of the functions are organized in their respective sub-modules, and the categories are taken from MetricsQL.
//! Any functions that cannot be categorized, will be placed directly within this module.
//! Note: PromQL's aggregate operators are provided as functions within the [`aggregate`] sub-module.
//!
//! Enable the `metricsql` feature flag to use MetricsQL specific functionality.
//! All the MetricsQL functions are prefixed by `mql_` to distinguish them from the PromQL functions,
//! as there is an overlap of functions betweent the two query languages. While MetricsQL is backwards
//! compatible with PromQL, it still extends the query language in many ways that isn't compatible with PromQL. Therefore, separation of the API will avoid any confusion and bugs.
//!
//! Unlike the general query builder API provided by this crate, the fns are provided as actual functions.
//! The returned types implement the `Display` and `IntoQuery` traits and work just like the types from the other parts of the API.
use super::{ops::Operable, IntoQuery};
use crate::seal::Sealed;
use std::{fmt, fmt::Display};

pub mod aggregate;
#[cfg(feature = "metricsql")]
pub mod label;
pub mod rollup;
#[cfg(feature = "metricsql")]
pub mod transform;

// This macro takes a name and formatting arguments which are directly passed to the `write` macro
// and constructs a QryFunc type. The name is generally provided directly as tokens (not as a quoted string).
// However, variables can be passed to the name field by specifying `var: var_name`
macro_rules! qry_fn {
    (var: $name: ident, { $($func: tt)+ }) => {
        QryFunc::new($name, move $($func)+)
    };
    (var: $name: ident, $($fmt: tt)+) => {
        QryFunc::new($name, move |f| write!(f, $($fmt)+))
    };
    ($name: tt, { $($func: tt)+ }) => {
        QryFunc::new(stringify!($name), move $($func)+)
    };
    ($name: tt, $($fmt: tt)+) => {
        QryFunc::new(stringify!($name), move |f| write!(f, $($fmt)+))
    };
}

pub(crate) use qry_fn;

/// Type that represents a query function. Meant to be constructed through the query functions provided in the [`fns`](crate::query::fns) module.
///
/// This internally holds an actual function (closure), that acts as a type that implements the `Display` trait.
#[derive(Debug, Clone)]
pub struct QryFunc<F: Fn(&mut fmt::Formatter) -> fmt::Result> {
    name: &'static str,
    args: F,
}

impl<F: Fn(&mut fmt::Formatter) -> fmt::Result> QryFunc<F> {
    fn new(name: &'static str, func: F) -> Self {
        Self { name, args: func }
    }
}

impl<F: Fn(&mut fmt::Formatter) -> fmt::Result> Display for QryFunc<F> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "{name}(", name = self.name)?; // write name and the starting parenthesis
        (self.args)(f)?; // write formatted arguments
        f.write_str(")") // closing parenthesis
    }
}

impl<F: Fn(&mut fmt::Formatter) -> fmt::Result> Sealed for QryFunc<F> {}

impl<F: Fn(&mut fmt::Formatter) -> fmt::Result> Operable for QryFunc<F> {}

impl<F: Fn(&mut fmt::Formatter) -> fmt::Result> IntoQuery for QryFunc<F> {
    type Target = String;
    fn into_query(self) -> Self::Target {
        self.to_string()
    }
}

#[inline]
fn basic_fn(
    name: &'static str,
    expr: impl Operable + 'static,
) -> QryFunc<impl Fn(&mut fmt::Formatter) -> fmt::Result + '_> {
    qry_fn!(var: name, "{expr}")
}

#[cfg(test)]
mod tests {
    use super::QryFunc;

    #[test]
    fn func_format() {
        let func = QryFunc::new("function", |f| f.write_str("arg1, arg2")).to_string();
        assert_eq!(func, "function(arg1, arg2)")
    }

    #[test]
    fn func_macro() {
        // Basic usage
        let func = qry_fn!(name, "args").to_string();
        assert_eq!(func, "name(args)");
        // Custom internal fn
        let func = qry_fn!(name, { |f| f.write_str("args") }).to_string();
        assert_eq!(func, "name(args)");
        // With Variables
        let x = "name";
        let func = qry_fn!(var: x, "args").to_string();
        assert_eq!(func, "name(args)");
        // Custom with Variables
        let x = "name";
        let func = qry_fn!(var: x, { |f| f.write_str("args") }).to_string();
        assert_eq!(func, "name(args)");
    }
}
